// Extension for exporting maps from "Tiled Map Editor" v.1.1 by 2023 (c) Dragons' Lord
// The map is saved in a text format that sjasmplus Cross-Assembler can read.

var acmeMapFormat = {
    name: "defb 00",
    extension: "asm",

    write: function(map, fileName) {
	
		// Create file
        var file = new TextFile(fileName, TextFile.WriteOnly);

		// Write each tile layer in map
        for (var i = 0; i < map.layerCount; ++i) {
            var layer = map.layerAt(i);
            if (layer.isTileLayer) {
				
                // Write each row in current layer
				for (y = 0; y < layer.height; ++y) {
                    var row = "\tdefb ";
                    
					// Write each column in current layer
					for (x = 0; x < (layer.width); ++x) {
						var cell = layer.cellAt(x, y);
						// Build a 8 bit hex value for each tile
						var val = cell.tileId+1;
						// let each value be in the format "xx"
						row += ("00" + val.toString(10)).slice(-2) + ",";
                    }
					row = row.slice(0,-1);
					file.writeLine(row);
                }
				file.writeLine("");
            }
        }
        file.commit();
    },
}

tiled.registerMapFormat("defb 00", acmeMapFormat)